#include <stdlib.h>
#include <stdio.h>
#include <sys/time.h>
#include <omp.h>
#include <signal.h>
#include <unistd.h>

//Prueba de Metodología de la Programación Paralela
//enero 2018
//Iteración de Jacobi vertical / diagonal, versión OpenMP 
//Esquema para entrada/salida, validación y toma de tiempos

void generar(int *m, int t,int lv,int uv) {
  int i;

  for (i = 0; i < t; i++) {
      m[i] =(int) (((1. * rand()) / RAND_MAX)*(uv-lv)+lv);
  }
}

void escribir(int *m, int N,int M) {
  int i, j;

  for (i = 0; i < N; i++) {
    for (j = 0; j < M; j++)
      printf("%d ", m[i * M + j]);
    printf("\n");
  }
  printf("\n");
}

void escribirresult(int *a,int N,int salida)
{
  int i;

  for(i=0;i<N;i++)
  {
        if((i%salida)==0)
        {
          printf("%d \n",a[i]);
        }
  }
  printf("\n");
}

/*
c
c     mseconds - returns elapsed milliseconds since Jan 1st, 1970.
c
*/
long long mseconds(){
  struct timeval t;
  gettimeofday(&t, NULL);
  return t.tv_sec*1000 + t.tv_usec/1000;
}

static void alarm_handler(int sig) {
  fprintf(stderr, "Time Limit Exceeded\n");
  abort();
}

extern void sec(int n,int *A,int numiter);

int main(int argc,char *argv[]) {
  int N; //tamaño matriz
  int *A;
  int cuantos; //número de problemas
  int numiter; //número de iteraciones
  int semilla; //semilla para generación aleatoria
  int lv,uv; //valores inferior y superior para generación aleatoria
  int salida; //indica cada cuantos elementos se escribe uno a la salida
  long long ti,tf,tt=0;

  FILE *stats_file = fopen("stats", "w");

  struct sigaction sact;
  sigemptyset(&sact.sa_mask);
  sact.sa_flags = 0;
  sact.sa_handler = alarm_handler;
  sigaction(SIGALRM, &sact, NULL);
  alarm(60);  /* time limit */
 
  scanf("%d",&cuantos);

  for(int i=0;i<cuantos;i++)
  {
      scanf("%d",&N);                   
      scanf("%d",&numiter);   
      scanf("%d",&semilla);  
      scanf("%d",&lv); 
      scanf("%d",&uv); 
      scanf("%d",&salida);  

      A = (int *) calloc(sizeof(int),N*N);

      srand(semilla);
      generar(A,N*N,lv,uv);
#ifdef DEBUG
  escribir(A,N,N);
#endif
    ti=mseconds(); 
    sec(N,A,numiter);
    tf=mseconds(); 
#ifdef DEBUG
  escribir(A,N,N);
#endif
      if(i!=0) tt+=tf-ti;
      escribirresult(A,N*N,salida);
      free(A);
  }
  
    fprintf(stats_file, "%Ld\n", tt);
    fclose(stats_file);
  return 0;
}
