/*
  CPP_CONTEST=2018Enero
  CPP_PROBLEM=C
  CPP_LANG=CPP+OPENMP+MPI
  CPP_PROCESSES_PER_NODE=calisto 2
*/

#include <stdlib.h>
#include <omp.h>
#include <mpi.h>

#define NUMTHREADS 2

extern void escribir(int *,int,int);

void filapar(int n,int *A,int *B)
{
  B[0]=(A[-n]+A[0]+A[1]+A[n])/5; //casilla izquierda
  for(int i=1;i<n-1;i++) //casillas no extremas de la fila
    B[i]=(A[i-n]+A[i-1]+A[i]+A[i+1]+A[i+n])/5;
  B[n-1]=(A[-1]+A[n-2]+A[n-1]+A[2*n-1])/5; //casilla derecha
}

void filaimpar(int n,int *A,int *B)
{
  B[0]=(A[-n+1]+A[0]+A[n+1])/5; //casilla izquierda
  for(int i=1;i<n-1;i++) //casillas no extremas de la fila
    B[i]=(A[i-n-1]+A[i-n+1]+A[i]+A[i+n-1]+A[i+n+1])/5;
  B[n-1]=(A[-2]+A[n-1]+A[2*n-2])/5; //casilla derecha
}

void par0(int nf,int n,int *A,int *B)
{
  B[n]=(A[n]+A[n+1]+A[2*n])/5; //casilla superior-izquierda
  for(int i=1;i<n-1;i++) //casillas no extremas de la primera fila
    B[n+i]=(A[n+i-1]+A[n+i]+A[n+i+1]+A[i+2*n])/5;
  B[2*n-1]=(A[2*n-2]+A[2*n-1]+A[3*n-1])/5; //casilla superior-derecha
#pragma omp parallel for
  for(int i=2;i<=nf;i++) 
    filapar(n,&A[i*n],&B[i*n]);
}

void parpm1(int nf,int n,int *A,int *B)
{
#pragma omp parallel for
  for(int i=1;i<nf;i++) 
    filapar(n,&A[i*n],&B[i*n]);
  B[nf*n]=(A[nf*n-n]+A[nf*n]+A[nf*n+1])/5; //casilla inferior-izquierda
  for(int i=1;i<n-1;i++) //casillas no extremas de la última fila
    B[nf*n+i]=(A[nf*n+i-n]+A[nf*n+i-1]+A[nf*n+i]+A[nf*n+i+1])/5;
  B[nf*n+n-1]=(A[nf*n-1]+A[nf*n+n-2]+A[nf*n+n-1])/5; //casilla inferior-derecha
}

void par(int nf,int n,int *A,int *B,int nodo,int np)
{
  if(nodo==0) //los nodos 0 y p-1 tratan distinto la primera y última fila
    par0(nf,n,A,B);
  else if(nodo==np-1)
    parpm1(nf,n,A,B);
  else
#pragma omp parallel for
    for(int i=1;i<=nf;i++) //se tratan todas las filas igual
      filapar(n,&A[i*n],&B[i*n]);
}

void impar0(int nf,int n,int *A,int *B)
{
  B[n]=(A[n]+A[2*n+1])/5; //casilla superior-izquierda
  for(int i=1;i<n-1;i++) //casillas no extremas de la primera fila
    B[n+i]=(A[2*n+i-1]+A[n+i]+A[2*n+i+1])/5;
  B[2*n-1]=(A[3*n-2]+A[2*n-1])/5; //casilla superior-derecha
#pragma omp parallel for
  for(int i=2;i<=nf;i++) 
    filaimpar(n,&A[i*n],&B[i*n]);
}

void imparpm1(int nf,int n,int *A,int *B)
{
#pragma omp parallel for
  for(int i=1;i<nf;i++) //para el resto de filas menos la última
    filaimpar(n,&A[i*n],&B[i*n]);
  B[nf*n]=(A[nf*n]+A[nf*n-n+1])/5; //casilla inferior-izquierda
  for(int i=1;i<n-1;i++) //casillas no extremas de la última fila
    B[nf*n+i]=(A[nf*n+i-n-1]+A[nf*n+i]+A[nf*n+i-n+1])/5;
  B[nf*n+n-1]=(A[nf*n-2]+A[nf*n+n-1])/5; //casilla inferior-derecha
}

void impar(int nf,int n,int *A,int *B,int nodo,int np)
{
   if(nodo==0)
    impar0(nf,n,A,B);
  else if(nodo==np-1)
    imparpm1(nf,n,A,B);
  else
#pragma omp parallel for
    for(int i=1;i<=nf;i++) 
      filaimpar(n,&A[i*n],&B[i*n]);
}

void copiar(int *vo,int n,int *vd)
{
  for(int i=0;i<n;i++)
    vd[i]=vo[i];
}

void ceros(int *a,int n)
{
  for(int i=0;i < n;i++)
    a[i]=0;
}

void comunicar(int n,int nf,int *A,int nodo,int np)
{
    MPI_Status *status;
    if(nodo!=0)
      MPI_Send(&A[n],n,MPI_INT,nodo-1,30,MPI_COMM_WORLD);
    if(nodo!=np-1)
      MPI_Send(&A[nf*n],n,MPI_INT,nodo+1,40,MPI_COMM_WORLD);
    if(nodo!=np-1)
      MPI_Recv(&A[(nf+1)*n],n,MPI_INT,nodo+1,30,MPI_COMM_WORLD,status);
    if(nodo!=0)
      MPI_Recv(A,n,MPI_INT,nodo-1,40,MPI_COMM_WORLD,status);
}

void  sec(int n,int *A,int numiter,int nodo,int np)
{
  MPI_Status *status;
  int *datos=new int[2]; //para almacenar n y numiter para mandarlos
  datos[0]=n;
  datos[1]=numiter;
  MPI_Bcast(datos,2,MPI_INT,0,MPI_COMM_WORLD);
  n=datos[0];
  numiter=datos[1];

  int *tamano=new int[np],*inicio=new int[np]; //arrays de tamaños e inicios, que calculan todos
  int tambas=n/np; //el tamaño básico
  for(int i=0;i<np;i++)
  {
    tamano[i]=tambas;
    if(i<n%np) //el resto de columnas asigna una más a los primeros procesos
      tamano[i]++;
  }
  inicio[0]=0;
  for(int i=1;i<np;i++)
  {
    inicio[i]=inicio[i-1]+tamano[i-1];
  }
  int nf=tamano[nodo]; //número de filas local

  int *Al=new int[(nf+2)*n]; //se reserva espacio para dos filas más para las filas que recibe de los vecinos
  int *Bl=new int[(nf+2)*n];

  if(nodo==0)
  {
    copiar(A,nf*n,&Al[n]); //para que el cero también tenga sus datos en la matriz local
    for(int i=1;i < np;i++)
    {
      MPI_Send(&A[inicio[i]*n],tamano[i]*n,MPI_INT,i,10,MPI_COMM_WORLD);
    }
  }
  else
      MPI_Recv(&Al[n],nf*n,MPI_INT,0,10,MPI_COMM_WORLD,status);
  ceros(Al,n); //pone ceros en la fila cero para tratar la primera fila como las demás
  ceros(&Al[(nf+1)*n],n); // y en la última fila 
  ceros(Bl,(nf+2)*n); //y en toda Bl

  int numiter2=numiter/2;

  omp_set_num_threads(NUMTHREADS);  
  for (int i = 0; i <numiter2; i++)
  {
    comunicar(n,nf,Al,nodo,np);
    par(nf,n,Al,Bl,nodo,np); //se añade un parámetro para el número de filas
#ifdef DEBUG
  escribir(Bl,nf+2,n);
#endif
    comunicar(n,nf,Bl,nodo,np);
    impar(nf,n,Bl,Al,nodo,np);
#ifdef DEBUG
  escribir(Al,nf+2,n);
#endif
  }
  if(numiter2*2!=numiter)
  {
    comunicar(n,nf,Al,nodo,np);
    par(nf,n,Al,Bl,nodo,np);
#ifdef DEBUG
  escribir(Bl,nf+2,n);
#endif
    copiar(&Bl[n],nf*n,&Al[n]);
  }
  
  if(nodo==0)
  {
    copiar(&Al[n],nf*n,A); //copia a la matriz global
    for(int i=1;i < np;i++)
      MPI_Recv(&A[inicio[i]*n],tamano[i]*n,MPI_INT,i,20,MPI_COMM_WORLD,status);
  }
  else
      MPI_Send(&Al[n],tamano[nodo]*n,MPI_INT,0,20,MPI_COMM_WORLD);

  delete[] inicio;
  delete[] tamano;
  delete[] datos;
  delete[] Al;
  delete[] Bl;
}

